/*
 * ExactPatternMatch.cpp
 *
 *  Created on: 9. 2. 2014
 *      Author: Jan Travnicek
 */

#include "ExactPatternMatch.h"
#include <tree/ranked/RankedTree.h>
#include <tree/ranked/RankedPattern.h>
#include <tree/ranked/RankedNonlinearPattern.h>
#include <tree/ranked/PrefixRankedTree.h>
#include <tree/ranked/PrefixRankedPattern.h>
#include <tree/ranked/PrefixRankedNonlinearPattern.h>
#include <tree/ranked/PrefixRankedBarTree.h>
#include <tree/ranked/PrefixRankedBarPattern.h>
#include <tree/ranked/PrefixRankedBarNonlinearPattern.h>
#include <tree/unranked/UnrankedTree.h>
#include <tree/unranked/UnrankedPattern.h>
#include "SubtreeJumpTable.h"
#include "ExactSubtreeRepeatsNaive.h"

#include <deque>
#include <foreach>

namespace arbology {

namespace exact {

std::set < unsigned > ExactPatternMatch::match ( const tree::Tree & subject, const tree::Tree & pattern ) {
	return dispatch ( subject.getData ( ), pattern.getData ( ) );
}

bool ExactPatternMatch::matchHelper ( const std::tree < alphabet::Symbol > & subject, const std::tree < alphabet::Symbol > & pattern, const alphabet::Symbol & subtreeVariable ) {
	if ( pattern.getData ( ) == subtreeVariable ) return true;

	if ( subject.getData ( ) != pattern.getData ( ) ) return false;

	if ( subject.getChildren ( ).size ( ) != pattern.getChildren ( ).size ( ) ) return false;

	for ( const std::tuple < const std::tree < alphabet::Symbol >, const std::tree < alphabet::Symbol > > & childs : std::make_tuple_foreach ( subject.getChildren ( ), pattern.getChildren ( ) ) )
		if ( !matchHelper ( std::get < 0 > ( childs ), std::get < 1 > ( childs ), subtreeVariable ) ) return false;

	return true;
}

bool ExactPatternMatch::matchHelper ( const std::tree < std::ranked_symbol < > > & subject, const std::tree < std::ranked_symbol < > > & pattern, const std::ranked_symbol < > & subtreeVariable ) {
	if ( pattern.getData ( ) == subtreeVariable ) return true;

	if ( subject.getData ( ) != pattern.getData ( ) ) return false;

	 // ranked symbols are the same; test for number of children is not needed
	for ( const std::tuple < const std::tree < std::ranked_symbol < > >, const std::tree < std::ranked_symbol < > > > & childs : std::make_tuple_foreach ( subject.getChildren ( ), pattern.getChildren ( ) ) )
		if ( !matchHelper ( std::get < 0 > ( childs ), std::get < 1 > ( childs ), subtreeVariable ) ) return false;

	return true;
}

bool ExactPatternMatch::matchHelper ( const std::tree < std::ranked_symbol < > > & subject, const std::tree < std::ranked_symbol < > > & pattern, const std::ranked_symbol < > & subtreeVariable, const std::set < std::ranked_symbol < > > & nonlinearVariables, const std::tree < std::ranked_symbol < > > & repeats, std::map < std::ranked_symbol < >, alphabet::Symbol > & variablesSetting ) {
	if ( pattern.getData ( ) == subtreeVariable ) return true;

	if ( nonlinearVariables.count ( pattern.getData ( ) ) ) {
		auto setting = variablesSetting.find ( pattern.getData ( ) );

		if ( setting != variablesSetting.end ( ) ) return repeats.getData ( ).getSymbol ( ) == setting->second;

		variablesSetting.insert ( std::make_pair ( pattern.getData ( ), repeats.getData ( ).getSymbol ( ) ) );

		return true;
	}

	if ( subject.getData ( ) != pattern.getData ( ) ) return false;

	 // ranked symbols are the same; test for number of children is not needed
	for ( const std::tuple < const std::tree < std::ranked_symbol < > >, const std::tree < std::ranked_symbol < > >, const std::tree < std::ranked_symbol < > > > & childs : std::make_tuple_foreach ( subject.getChildren ( ), pattern.getChildren ( ), repeats.getChildren ( ) ) )
		if ( !matchHelper ( std::get < 0 > ( childs ), std::get < 1 > ( childs ), subtreeVariable, nonlinearVariables, std::get < 2 > ( childs ), variablesSetting ) ) return false;

	return true;
}

void ExactPatternMatch::matchInternal ( unsigned & index, std::set < unsigned > & occ, const std::tree < alphabet::Symbol > & subject, const std::tree < alphabet::Symbol > & pattern, const alphabet::Symbol & subtreeVariable ) {
	if ( matchHelper ( subject, pattern, subtreeVariable ) ) occ.insert ( index );

	index++;

	for ( const std::tree < alphabet::Symbol > & child : subject.getChildren ( ) )
		matchInternal ( index, occ, child, pattern, subtreeVariable );
}

void ExactPatternMatch::matchInternal ( unsigned & index, std::set < unsigned > & occ, const std::tree < std::ranked_symbol < > > & subject, const std::tree < std::ranked_symbol < > > & pattern, const std::ranked_symbol < > & subtreeVariable ) {
	if ( matchHelper ( subject, pattern, subtreeVariable ) ) occ.insert ( index );

	index++;

	for ( const std::tree < std::ranked_symbol < > > & child : subject.getChildren ( ) )
		matchInternal ( index, occ, child, pattern, subtreeVariable );
}

void ExactPatternMatch::matchInternal ( unsigned & index, std::set < unsigned > & occ, const std::tree < std::ranked_symbol < > > & subject, const std::tree < std::ranked_symbol < > > & pattern, const std::ranked_symbol < > & subtreeVariable, const std::set < std::ranked_symbol < > > & nonlinearVariables, const std::tree < std::ranked_symbol < > > & repeats ) {
	std::map < std::ranked_symbol < >, alphabet::Symbol > variablesSetting;

	if ( matchHelper ( subject, pattern, subtreeVariable, nonlinearVariables, repeats, variablesSetting ) ) occ.insert ( index );

	index++;

	for ( const std::tuple < const std::tree < std::ranked_symbol < > >, const std::tree < std::ranked_symbol < > > > & childs : std::make_tuple_foreach ( subject.getChildren ( ), repeats.getChildren ( ) ) )
		matchInternal ( index, occ, std::get < 0 > ( childs ), pattern, subtreeVariable, nonlinearVariables, std::get < 1 > ( childs ) );
}

std::set < unsigned > ExactPatternMatch::match ( const tree::UnrankedTree < > & subject, const tree::UnrankedPattern < > & pattern ) {
	unsigned i = 0;
	std::set < unsigned > occ;

	matchInternal ( i, occ, subject.getContent ( ), pattern.getContent ( ), pattern.getSubtreeWildcard ( ) );
	return occ;
}

auto ExactPatternMatchUnrankedPattern = ExactPatternMatch::RegistratorWrapper < std::set < unsigned >, tree::UnrankedTree < >, tree::UnrankedPattern < > > ( ExactPatternMatch::match );

std::set < unsigned > ExactPatternMatch::match ( const tree::RankedTree < > & subject, const tree::RankedPattern < > & pattern ) {
	unsigned i = 0;
	std::set < unsigned > occ;

	matchInternal ( i, occ, subject.getContent ( ), pattern.getContent ( ), pattern.getSubtreeWildcard ( ) );
	return occ;
}

auto ExactPatternMatchRankedPattern = ExactPatternMatch::RegistratorWrapper < std::set < unsigned >, tree::RankedTree < >, tree::RankedPattern < > > ( ExactPatternMatch::match );

std::set < unsigned > ExactPatternMatch::match ( const tree::RankedTree < > & subject, const tree::RankedNonlinearPattern < > & pattern ) {
	unsigned i = 0;
	std::set < unsigned > occ;

	tree::RankedTree < > repeats = ExactSubtreeRepeatsNaive::repeats ( subject );

	matchInternal ( i, occ, subject.getContent ( ), pattern.getContent ( ), pattern.getSubtreeWildcard ( ), pattern.getNonlinearVariables ( ), repeats.getContent ( ) );
	return occ;
}

auto ExactPatternMatchRankedNonlinearPattern = ExactPatternMatch::RegistratorWrapper < std::set < unsigned >, tree::RankedTree < >, tree::RankedNonlinearPattern < > > ( ExactPatternMatch::match );

std::set < unsigned > ExactPatternMatch::match ( const tree::PrefixRankedTree < > & subject, const tree::PrefixRankedPattern < > & pattern ) {
	std::vector < int > subjectSubtreeJumpTable = SubtreeJumpTable::compute ( subject );

	std::set < unsigned > occ;

	for ( unsigned i = 0; i + pattern.getContent ( ).size ( ) <= subject.getContent ( ).size ( ); i++ ) {
		unsigned j = 0;
		unsigned offset = i;

		for ( ; j < pattern.getContent ( ).size ( ); j++ ) {
			if ( pattern.getContent ( )[j] == subject.getContent ( )[offset] )
				offset++;
			else if ( pattern.getContent ( )[j] == pattern.getSubtreeWildcard ( ) )
				offset = subjectSubtreeJumpTable[offset];
			else
				break;
		}

		if ( j == pattern.getContent ( ).size ( ) )
			occ.insert ( i );
	}

	return occ;
}

auto ExactPatternMatchPrefixRankedPattern = ExactPatternMatch::RegistratorWrapper < std::set < unsigned >, tree::PrefixRankedTree < >, tree::PrefixRankedPattern < > > ( ExactPatternMatch::match );

std::set < unsigned > ExactPatternMatch::match ( const tree::PrefixRankedTree < > & subject, const tree::PrefixRankedNonlinearPattern < > & pattern ) {
	std::vector < int > subjectSubtreeJumpTable = SubtreeJumpTable::compute ( subject );
	std::map < std::ranked_symbol < >, alphabet::Symbol > variablesSetting;

	tree::PrefixRankedTree < > repeats = ExactSubtreeRepeatsNaive::repeats ( subject );

	std::set < unsigned > occ;

	for ( unsigned i = 0; i + pattern.getContent ( ).size ( ) <= subject.getContent ( ).size ( ); i++ ) {
		variablesSetting.clear();
		unsigned j = 0;
		unsigned offset = i;

		for ( ; j < pattern.getContent ( ).size ( ); j++ ) {
			if ( pattern.getContent ( )[j] == subject.getContent ( )[offset] )
				offset++;
			else if ( pattern.getContent ( )[j] == pattern.getSubtreeWildcard ( ) )
				offset = subjectSubtreeJumpTable[offset];
			else if ( pattern.getNonlinearVariables ( ).count ( pattern.getContent ( )[j] ) ) {
				auto setting = variablesSetting.find ( pattern.getContent ( )[j] );

				if ( setting != variablesSetting.end ( ) && repeats.getContent ( )[offset].getSymbol ( ) != setting->second)
					break;

				variablesSetting.insert ( std::make_pair ( pattern.getContent ( )[j], repeats.getContent( )[offset].getSymbol ( ) ) );
				offset = subjectSubtreeJumpTable[offset];
			} else
				break;
		}

		if ( j == pattern.getContent ( ).size ( ) )
			occ.insert ( i );
	}

	return occ;
}

auto ExactPatternMatchPrefixRankedNonlinearPattern = ExactPatternMatch::RegistratorWrapper < std::set < unsigned >, tree::PrefixRankedTree < >, tree::PrefixRankedNonlinearPattern < > > ( ExactPatternMatch::match );

std::set < unsigned > ExactPatternMatch::match ( const tree::PrefixRankedBarTree < > & subject, const tree::PrefixRankedBarPattern < > & pattern ) {
	std::vector < int > subjectSubtreeJumpTable = SubtreeJumpTable::compute ( subject );

	std::set < unsigned > occ;

	for ( unsigned i = 0; i + pattern.getContent ( ).size ( ) <= subject.getContent ( ).size ( ); i++ ) {
		unsigned j = 0;
		unsigned offset = i;

		for ( ; j < pattern.getContent ( ).size ( ); j++ ) {
			if ( pattern.getContent ( )[j] == subject.getContent ( )[offset] ) {
				offset++;
			} else if ( pattern.getContent ( )[j] == pattern.getSubtreeWildcard ( ) ) {
				j++;
				offset = subjectSubtreeJumpTable[offset];
			} else {
				break;
			}
		}

		if ( j == pattern.getContent ( ).size ( ) )
			occ.insert ( i );
	}

	return occ;
}

auto ExactPatternMatchPrefixRankedBarPattern = ExactPatternMatch::RegistratorWrapper < std::set < unsigned >, tree::PrefixRankedBarTree < >, tree::PrefixRankedBarPattern < > > ( ExactPatternMatch::match );

std::set < unsigned > ExactPatternMatch::match ( const tree::PrefixRankedBarTree < > & subject, const tree::PrefixRankedBarNonlinearPattern < > & pattern ) {
	std::vector < int > subjectSubtreeJumpTable = SubtreeJumpTable::compute ( subject );
	std::map < std::ranked_symbol < >, alphabet::Symbol > variablesSetting;

	tree::PrefixRankedBarTree < > repeats = ExactSubtreeRepeatsNaive::repeats ( subject );

	std::set < unsigned > occ;

	for ( unsigned i = 0; i + pattern.getContent ( ).size ( ) <= subject.getContent ( ).size ( ); i++ ) {
		variablesSetting.clear();
		unsigned j = 0;
		unsigned offset = i;

		for ( ; j < pattern.getContent ( ).size ( ); j++ ) {
			if ( pattern.getContent ( )[j] == subject.getContent ( )[offset] ) {
				offset++;
			} else if ( pattern.getContent ( )[j] == pattern.getSubtreeWildcard ( ) ) {
				j++;
				offset = subjectSubtreeJumpTable[offset];
			} else if ( pattern.getNonlinearVariables ( ).count ( pattern.getContent ( )[j] ) ) {
				auto setting = variablesSetting.find ( pattern.getContent ( )[j] );

				if ( setting != variablesSetting.end ( ) && repeats.getContent ( )[offset].getSymbol ( ) != setting->second)
					break;

				variablesSetting.insert ( std::make_pair ( pattern.getContent ( )[j], repeats.getContent( )[offset].getSymbol ( ) ) );

				j++;
				offset = subjectSubtreeJumpTable[offset];
			} else {
				break;
			}
		}

		if ( j == pattern.getContent ( ).size ( ) )
			occ.insert ( i );
	}

	return occ;
}

auto ExactPatternMatchPrefixRankedBarNonlinearPattern = ExactPatternMatch::RegistratorWrapper < std::set < unsigned >, tree::PrefixRankedBarTree < >, tree::PrefixRankedBarNonlinearPattern < > > ( ExactPatternMatch::match );

} /* namespace exact */

} /* namespace arbology */
