/*
 * EpsilonClosure.cpp
 *
 *  Created on: 23. 3. 2014
 *	  Author: Tomas Pecka
 */

#include "EpsilonClosure.h"

#include <automaton/FSM/ExtendedNFA.h>
#include <automaton/FSM/CompactNFA.h>
#include <automaton/FSM/EpsilonNFA.h>
#include <automaton/FSM/MultiInitialStateNFA.h>
#include <automaton/FSM/NFA.h>
#include <automaton/FSM/DFA.h>

#include <set>
#include <map>
#include <queue>

#include "../../regexp/properties/RegExpEpsilon.h"

namespace automaton {

namespace properties {

std::set<label::Label> EpsilonClosure::epsilonClosure( const automaton::EpsilonNFA < > & fsm, const label::Label & q ) {
	if(! fsm.getStates().count(q) ) throw exception::CommonException("State is not in the automaton");

	std::set<label::Label> closure;
	std::queue<label::Label> queue;
	std::map<label::Label, bool> visited;

	for( const auto & p : fsm.getStates( ) )
		visited[ p ] = false;

	queue.push( q );
	while( ! queue.empty( ) )
	{
		label::Label p = queue.front( );
		queue.pop( );
		visited[ p ] = true;
		closure.insert( p );

		auto tos = fsm.getTransitions( ).find(std::make_pair(p, std::variant<string::Epsilon < >, alphabet::Symbol>(string::Epsilon < >::EPSILON) ) );
		if(tos == fsm.getTransitions().end()) continue;

		for( const auto & to : tos->second )
			if( visited [ to ] == false )
				queue.push( to );
	}

	return closure;
}

auto EpsilonClosureEpsilonNFA = EpsilonClosure::RegistratorWrapper<std::set<label::Label>, automaton::EpsilonNFA < > >(EpsilonClosure::epsilonClosure);

std::set<label::Label> EpsilonClosure::epsilonClosure( const automaton::MultiInitialStateNFA < > & fsm, const label::Label & q ) {
	if(! fsm.getStates().count(q) ) throw exception::CommonException("State is not in the automaton");

	std::set<label::Label> closure;
	closure.insert(q);
	return closure;
}

auto EpsilonClosureMultiInitialStateNFA = EpsilonClosure::RegistratorWrapper<std::set<label::Label>, automaton::MultiInitialStateNFA < > >(EpsilonClosure::epsilonClosure);

std::set<label::Label> EpsilonClosure::epsilonClosure( const automaton::NFA < > & fsm, const label::Label & q ) {
	if(! fsm.getStates().count(q) ) throw exception::CommonException("State is not in the automaton");

	std::set<label::Label> closure;
	closure.insert(q);
	return closure;
}

auto EpsilonClosureNFA = EpsilonClosure::RegistratorWrapper<std::set<label::Label>, automaton::NFA < > >(EpsilonClosure::epsilonClosure);

std::set<label::Label> EpsilonClosure::epsilonClosure( const automaton::DFA < > & fsm, const label::Label & q ) {
	if(! fsm.getStates().count(q) ) throw exception::CommonException("State is not in the automaton");

	std::set<label::Label> closure;
	closure.insert(q);
	return closure;
}

auto EpsilonClosureDFA = EpsilonClosure::RegistratorWrapper<std::set<label::Label>, automaton::DFA<>>(EpsilonClosure::epsilonClosure);

std::set<label::Label> EpsilonClosure::epsilonClosure( const automaton::ExtendedNFA & fsm, const label::Label & q ) {
	if(! fsm.getStates().count(q) ) throw exception::CommonException("State is not in the automaton");

	std::set<label::Label> closure;
	std::queue<label::Label> queue;
	std::map<label::Label, bool> visited;

	for( const auto & p : fsm.getStates( ) )
		visited[ p ] = false;

	queue.push( q );
	while( ! queue.empty( ) )
	{
		label::Label p = queue.front( );
		queue.pop( );
		visited[ p ] = true;
		closure.insert( p );

		for( const auto & transition : fsm.getTransitionsFromState( p ) )
			if( regexp::properties::RegExpEpsilon::languageContainsEpsilon( transition.first.second ) )
				for (const auto & to : transition.second )
					if( visited [ to ] == false )
						queue.push( to );
	}

	return closure;
}

auto EpsilonClosureExtendedNFA = EpsilonClosure::RegistratorWrapper<std::set<label::Label>, automaton::ExtendedNFA>(EpsilonClosure::epsilonClosure);

std::set<label::Label> EpsilonClosure::epsilonClosure( const automaton::CompactNFA < > & fsm, const label::Label & q ) {
	if(! fsm.getStates().count(q) ) throw exception::CommonException("State is not in the automaton");

	std::set<label::Label> closure;
	std::queue<label::Label> queue;
	std::map<label::Label, bool> visited;

	for( const auto & p : fsm.getStates( ) )
		visited[ p ] = false;

	queue.push( q );
	while( ! queue.empty( ) )
	{
		label::Label p = queue.front( );
		queue.pop( );
		visited[ p ] = true;
		closure.insert( p );

		for( const auto & transition : fsm.getTransitionsFromState( p ) )
			if( transition.first.second.size() == 0 )
				for (const auto & to : transition.second )
					if( visited [ to ] == false )
						queue.push( to );
	}

	return closure;
}

auto EpsilonClosureCompactNFA = EpsilonClosure::RegistratorWrapper<std::set<label::Label>, automaton::CompactNFA < > >(EpsilonClosure::epsilonClosure);

std::set<label::Label> EpsilonClosure::epsilonClosure(const Automaton& automaton, const label::Label& q) {
	return dispatch(automaton.getData(), q);
}

} /* namespace properties */

} /* namespace automaton */

