/*
 * Pair.h
 *
 * Created on: Apr 1, 2013
 * Author: Jan Travnicek
 */

#ifndef OBJECTS_PAIR_H_
#define OBJECTS_PAIR_H_

#include <pair>
#include <string>
#include <sstream>

#include "ContainerBase.h"
#include "ContainerFeatures.h"

#include <core/xmlApi.hpp>

namespace container {

/**
 * Basic container from which are derived all other containers.
 * Contains reason why the container occured.
 */
template < class KeyType, class ValueType >
class ObjectsPair : public std::pair < KeyType, ValueType >, public ContainerBase {
public:
	explicit ObjectsPair ( alib::Object first, alib::Object second );
	explicit ObjectsPair ( std::pair < KeyType, ValueType > );

	virtual ContainerBase * clone ( ) const;

	virtual ContainerBase * plunder ( ) &&;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const ObjectsPair & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "Pair";

		return xmlTagName;
	}

	static ObjectsPair parse ( std::deque < sax::Token >::iterator & input );
	static std::pair < KeyType, ValueType > parseRaw ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;
	static void compose ( std::deque < sax::Token > & out, const std::pair < KeyType, ValueType > & input );
};

template < class KeyType, class ValueType >
ObjectsPair < KeyType, ValueType >::ObjectsPair(std::pair<KeyType, ValueType> pair) : std::pair<KeyType, ValueType>(std::move(pair)) {

}

template < class KeyType, class ValueType >
ObjectsPair < KeyType, ValueType >::ObjectsPair(alib::Object first, alib::Object second) : std::pair<KeyType, ValueType>(std::move(first), std::move(second)) {

}

template < class KeyType, class ValueType >
ContainerBase* ObjectsPair < KeyType, ValueType >::clone() const {
	return new ObjectsPair(*this);
}

template < class KeyType, class ValueType >
ContainerBase* ObjectsPair < KeyType, ValueType >::plunder() && {
	return new ObjectsPair(std::move(*this));
}

template < class KeyType, class ValueType >
int ObjectsPair < KeyType, ValueType >::compare(const ObjectsPair& other) const {
	std::compare<std::pair<KeyType, ValueType>> comp;
	return comp(static_cast<const std::pair<KeyType, ValueType>>(*this), static_cast<const std::pair<KeyType, ValueType>>(other));
}

template < class KeyType, class ValueType >
void ObjectsPair < KeyType, ValueType >::operator>>(std::ostream& os) const {
	os << "(ObjectsPair " << static_cast<const std::pair<KeyType, ValueType>>(*this) << ")";
}

template < class KeyType, class ValueType >
ObjectsPair < KeyType, ValueType >::operator std::string() const {
	std::stringstream ss;
	ss << *this;
	return std::move(ss).str();
}

template < class KeyType, class ValueType >
ObjectsPair < KeyType, ValueType > ObjectsPair < KeyType, ValueType >::parse(std::deque<sax::Token>::iterator& input) {
	return ObjectsPair < KeyType, ValueType > ( parseRaw ( input ) );
}

template < class KeyType, class ValueType >
void ObjectsPair < KeyType, ValueType >::compose(std::deque<sax::Token>& out) const {
	compose(out, *this);
}

template < class KeyType, class ValueType >
std::pair < KeyType, ValueType > ObjectsPair < KeyType, ValueType >::parseRaw ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, ObjectsPair::getXmlTagName() );

	KeyType first = alib::xmlApi < KeyType >::parse ( input );
	ValueType second = alib::xmlApi < ValueType >::parse ( input );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, ObjectsPair::getXmlTagName() );

	return std::make_pair ( std::move ( first ), std::move ( second ) );
}

template < class KeyType, class ValueType >
void ObjectsPair < KeyType, ValueType >::compose ( std::deque < sax::Token > & out, const std::pair < KeyType, ValueType > & input ) {
	out.emplace_back ( ObjectsPair::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );

	alib::xmlApi < KeyType >::compose ( out, input.first );
	alib::xmlApi < ValueType >::compose ( out, input.second );

	out.emplace_back ( ObjectsPair::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

} /* namespace container */

namespace alib {

template < typename T, typename R >
struct xmlApi < std::pair < T, R > > {
	static std::pair < T, R > parse ( std::deque < sax::Token >::iterator & input );
	static bool first ( const std::deque < sax::Token >::const_iterator & input );
	static void compose ( std::deque < sax::Token > & output, const std::pair < T, R > & data );
};

template < typename T, typename R >
std::pair < T, R > xmlApi < std::pair < T, R > >::parse ( std::deque < sax::Token >::iterator & input ) {
	return container::ObjectsPair < T, R >::parseRaw ( input );
}

template < typename T, typename R >
bool xmlApi < std::pair < T, R > >::first ( const std::deque < sax::Token >::const_iterator & input ) {
	return sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, container::ObjectsPair < >::getXmlTagName() );
}

template < typename T, typename R >
void xmlApi < std::pair < T, R > >::compose ( std::deque < sax::Token > & output, const std::pair < T, R > & input ) {
	return container::ObjectsPair < T, R >::compose ( output, input );
}

} /* namespace alib */

#endif /* OBJECTS_PAIR_H_ */
