/*
 * ExtendedNFA.cpp
 *
 *  Created on: Mar 25, 2013
 *      Author: Jan Travnicek
 */

#include "ExtendedNFA.h"
#include "CompactNFA.h"
#include "EpsilonNFA.h"
#include "MultiInitialStateNFA.h"
#include "NFA.h"
#include "DFA.h"
#include "../../regexp/RegExp.h"
#include <ostream>
#include <algorithm>
#include <sstream>

#include <sax/FromXMLParserHelper.h>
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"
#include "../Automaton.h"
#include <object/Object.h>
#include <core/xmlApi.hpp>
#include <core/castApi.hpp>
#include "../../label/InitialStateLabel.h"

namespace automaton {

ExtendedNFA::ExtendedNFA ( std::set < label::Label > states, std::set < alphabet::Symbol > inputAlphabet, label::Label initialState, std::set < label::Label > finalStates ) : std::Components < ExtendedNFA, alphabet::Symbol, std::tuple < InputAlphabet >, std::tuple < >, label::Label, std::tuple < States, FinalStates >, std::tuple < InitialState > > ( std::make_tuple ( std::move ( inputAlphabet ) ), std::tuple < > ( ), std::make_tuple ( std::move ( states ), std::move ( finalStates ) ), std::make_tuple ( std::move ( initialState ) ) ) {
}

ExtendedNFA::ExtendedNFA ( label::Label initialState ) : ExtendedNFA ( std::set < label::Label > { initialState }, std::set < alphabet::Symbol > { }, initialState, std::set < label::Label > { } ) {
}

ExtendedNFA::ExtendedNFA ( const CompactNFA < > & other ) : ExtendedNFA ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialState ( ), other.getFinalStates ( ) ) {
	for ( const auto & transition : other.getTransitions ( ) ) {
		std::pair < label::Label, regexp::RegExp > key = std::make_pair ( transition.first.first, regexp::RegExp ( regexp::regexpFrom ( transition.first.second ) ) );
		transitions[key] = transition.second;
	}
}

ExtendedNFA::ExtendedNFA ( const EpsilonNFA < > & other ) : ExtendedNFA ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialState ( ), other.getFinalStates ( ) ) {
	for ( const auto & transition : other.getTransitions ( ) ) {
		if ( transition.first.second.is < string::Epsilon < > > ( ) ) {
			std::pair < label::Label, regexp::RegExp > key = std::make_pair ( transition.first.first, regexp::RegExp ( regexp::regexpFrom ( ) ) );
			transitions[key] = transition.second;
		} else {
			std::pair < label::Label, regexp::RegExp > key = std::make_pair ( transition.first.first, regexp::RegExp ( regexp::regexpFrom ( transition.first.second.get < alphabet::Symbol > ( ) ) ) );
			transitions[key] = transition.second;
		}
	}
}

ExtendedNFA::ExtendedNFA ( const MultiInitialStateNFA < > & other ) : ExtendedNFA ( other.getStates ( ) + std::set < label::Label > { label::createUniqueLabel ( label::InitialStateLabel::INITIAL_STATE_LABEL, other.getStates ( ) ) }, other.getInputAlphabet ( ), label::createUniqueLabel ( label::InitialStateLabel::INITIAL_STATE_LABEL, other.getStates ( ) ), other.getFinalStates ( ) ) {
	for ( const auto & transition : other.getTransitions ( ) ) {
		std::pair < label::Label, regexp::RegExp > key = std::make_pair ( transition.first.first, regexp::RegExp ( regexp::regexpFrom ( transition.first.second ) ) );
		transitions[key] = transition.second;
	}

	std::pair < label::Label, regexp::RegExp > key = std::make_pair ( this->getInitialState ( ), regexp::RegExp ( regexp::regexpFrom ( ) ) );
	transitions[key] = other.getInitialStates ( );
}

ExtendedNFA::ExtendedNFA ( const NFA <> & other ) : ExtendedNFA ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialState ( ), other.getFinalStates ( ) ) {
	for ( const auto & transition : other.getTransitions ( ) ) {
		std::pair < label::Label, regexp::RegExp > key = std::make_pair ( transition.first.first, regexp::RegExp ( regexp::regexpFrom ( transition.first.second ) ) );
		transitions[key] = transition.second;
	}
}

ExtendedNFA::ExtendedNFA ( const DFA<> & other ) : ExtendedNFA ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialState ( ), other.getFinalStates ( ) ) {
	for ( const auto & transition : other.getTransitions ( ) ) {
		std::pair < label::Label, regexp::RegExp > key = std::make_pair ( transition.first.first, regexp::RegExp ( regexp::regexpFrom ( transition.first.second ) ) );
		transitions[key].insert ( transition.second );
	}
}

AutomatonBase * ExtendedNFA::clone ( ) const {
	return new ExtendedNFA ( * this );
}

AutomatonBase * ExtendedNFA::plunder ( ) && {
	return new ExtendedNFA ( std::move ( * this ) );
}

bool ExtendedNFA::addTransition ( label::Label from, regexp::RegExp input, label::Label to ) {
	if ( !getStates ( ).count ( from ) )
		throw AutomatonException ( "State \"" + std::to_string ( from ) + "\" doesn't exist." );

	const std::set < alphabet::Symbol > & inputRegExpAlphabet = input.getAlphabet ( );

	 // Transition regexp's alphabet must be subset of automaton's alphabet
	if ( !std::includes ( getInputAlphabet ( ).begin ( ), getInputAlphabet ( ).end ( ), inputRegExpAlphabet.begin ( ), inputRegExpAlphabet.end ( ) ) )
		throw AutomatonException ( "Input string is over different alphabet than automaton" );

	if ( !getStates ( ).count ( to ) )
		throw AutomatonException ( "State \"" + std::to_string ( to ) + "\" doesn't exist." );

	std::pair < label::Label, regexp::RegExp > key = std::make_pair ( std::move ( from ), std::move ( input ) );

	return transitions[std::move ( key )].insert ( std::move ( to ) ).second;
}

bool ExtendedNFA::removeTransition ( const label::Label & from, const regexp::RegExp & input, const label::Label & to ) {
	std::pair < label::Label, regexp::RegExp > key = std::make_pair ( from, input );

	return transitions[key].erase ( to );
}

const std::map < std::pair < label::Label, regexp::RegExp >, std::set < label::Label > > & ExtendedNFA::getTransitions ( ) const {
	return transitions;
}

std::map < std::pair < label::Label, regexp::RegExp >, std::set < label::Label > > ExtendedNFA::getTransitionsFromState ( const label::Label & from ) const {
	if ( !getStates ( ).count ( from ) )
		throw AutomatonException ( "State \"" + std::to_string ( from ) + "\" doesn't exist" );

	std::map < std::pair < label::Label, regexp::RegExp >, std::set < label::Label > > transitionsFromState;

	for ( const std::pair < const std::pair < label::Label, regexp::RegExp >, std::set < label::Label > > & transition : transitions )
		if ( transition.first.first == from )
			transitionsFromState.insert ( make_pair ( transition.first, transition.second ) );

	return transitionsFromState;
}

std::map < std::pair < label::Label, regexp::RegExp >, std::set < label::Label > > ExtendedNFA::getTransitionsToState ( const label::Label & to ) const {
	if ( !getStates ( ).count ( to ) )
		throw AutomatonException ( "State \"" + std::to_string ( to ) + "\" doesn't exist" );

	std::map < std::pair < label::Label, regexp::RegExp >, std::set < label::Label > > transitionsToState;

	for ( const std::pair < const std::pair < label::Label, regexp::RegExp >, std::set < label::Label > > & transition : transitions )
		if ( transition.second.find ( to ) != transition.second.end ( ) )
			transitionsToState.insert ( make_pair ( transition.first, transition.second ) );

	return transitionsToState;
}

int ExtendedNFA::compare ( const ExtendedNFA & other ) const {
	auto first = std::tie ( getStates ( ), getInputAlphabet ( ), getInitialState ( ), getFinalStates ( ), transitions );
	auto second = std::tie ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialState ( ), other.getFinalStates ( ), other.getTransitions ( ) );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

void ExtendedNFA::operator >>( std::ostream & out ) const {
	out << "(ExtendedNFA "
	    << "states = " << getStates ( )
	    << "inputAlphabet = " << getInputAlphabet ( )
	    << "initialState = " << getInitialState ( )
	    << "finalStates = " << getFinalStates ( )
	    << "transitions = " << transitions
	    << ")";
}

ExtendedNFA::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

ExtendedNFA ExtendedNFA::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, ExtendedNFA::getXmlTagName() );

	std::set < label::Label > states = AutomatonFromXMLParser::parseStates < label::Label > ( input );
	std::set < alphabet::Symbol > inputSymbols = AutomatonFromXMLParser::parseInputAlphabet < alphabet::Symbol > ( input );
	label::Label initialState = AutomatonFromXMLParser::parseInitialState < label::Label > ( input );
	std::set < label::Label > finalStates = AutomatonFromXMLParser::parseFinalStates < label::Label > ( input );

	ExtendedNFA automaton ( std::move ( initialState ) );

	automaton.setStates ( std::move ( states ) );
	automaton.setInputAlphabet ( std::move ( inputSymbols ) );
	automaton.setFinalStates ( std::move ( finalStates ) );

	AutomatonFromXMLParser::parseTransitions < ExtendedNFA > ( input, automaton );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, ExtendedNFA::getXmlTagName() );
	return automaton;
}

void ExtendedNFA::parseTransition ( std::deque < sax::Token >::iterator & input, ExtendedNFA & automaton ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "transition" );
	label::Label from = AutomatonFromXMLParser::parseTransitionFrom < label::Label > ( input );
	regexp::RegExp inputRegexp = AutomatonFromXMLParser::parseTransitionInputRegexp ( input );
	label::Label to = AutomatonFromXMLParser::parseTransitionTo < label::Label > ( input );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "transition" );

	automaton.addTransition ( std::move ( from ), std::move ( inputRegexp ), std::move ( to ) );
}

void ExtendedNFA::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( ExtendedNFA::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );

	AutomatonToXMLComposer::composeStates ( out, this->getStates ( ) );
	AutomatonToXMLComposer::composeInputAlphabet ( out, this->getInputAlphabet ( ) );
	AutomatonToXMLComposer::composeInitialState ( out, this->getInitialState ( ) );
	AutomatonToXMLComposer::composeFinalStates ( out, this->getFinalStates ( ) );
	composeTransitions ( out );

	out.emplace_back ( ExtendedNFA::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

void ExtendedNFA::composeTransitions ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( "transitions", sax::Token::TokenType::START_ELEMENT );

	for ( const auto & transition : this->getTransitions ( ) )
		for ( const auto & targetState : transition.second ) {
			out.emplace_back ( "transition", sax::Token::TokenType::START_ELEMENT );

			AutomatonToXMLComposer::composeTransitionFrom ( out, transition.first.first );
			AutomatonToXMLComposer::composeTransitionInputRegexp ( out, transition.first.second );
			AutomatonToXMLComposer::composeTransitionTo ( out, targetState );

			out.emplace_back ( "transition", sax::Token::TokenType::END_ELEMENT );
		}

	out.emplace_back ( "transitions", sax::Token::TokenType::END_ELEMENT );
}

} /* namespace automaton */

namespace alib {

auto extendedNFAParserRegister = xmlApi < automaton::Automaton >::ParserRegister < automaton::ExtendedNFA > ( );
auto extendedNFAParserRegister2 = xmlApi < alib::Object >::ParserRegister < automaton::ExtendedNFA > ( );

auto ExtendedNFAFromDFA = castApi::CastRegister < automaton::ExtendedNFA, automaton::DFA < > > ( );
auto ExtendedNFAFromNFA = castApi::CastRegister < automaton::ExtendedNFA, automaton::NFA < > > ( );
auto ExtendedNFAFromMultiInitialStateNFA = castApi::CastRegister < automaton::ExtendedNFA, automaton::MultiInitialStateNFA < > > ( );
auto ExtendedNFAEpsilonNFA = castApi::CastRegister < automaton::ExtendedNFA, automaton::EpsilonNFA < > > ( );
auto ExtendedNFACompactNFA = castApi::CastRegister < automaton::ExtendedNFA, automaton::CompactNFA < > > ( );
auto ExtendedNFACastBinder = castApi::CastPoolStringBinder < automaton::ExtendedNFA > ( automaton::ExtendedNFA::getXmlTagName() );

} /* namespace alib */
