/*
 * ExtendedNFA.h
 *
 *  Created on: Mar 25, 2013
 *      Author: Jan Travnicek
 */

#ifndef EXTENDED_NFA_H_
#define EXTENDED_NFA_H_

#include "../AutomatonException.h"
#include <map>
#include <core/components.hpp>
#include "../AutomatonBase.h"
#include "../../regexp/RegExp.h"
#include "../../label/Label.h"
#include "CompactNFA.h"
#include "EpsilonNFA.h"
#include "MultiInitialStateNFA.h"
#include "NFA.h"
#include "DFA.h"

namespace automaton {

/**
 * Represents Finite Automaton.
 * Can store nondeterministic finite automaton without epsilon transitions.
 */
class InputAlphabet;
class States;
class FinalStates;
class InitialState;

class ExtendedNFA : public AutomatonBase, public std::Components < ExtendedNFA, alphabet::Symbol, std::tuple < InputAlphabet >, std::tuple < >, label::Label, std::tuple < States, FinalStates >, std::tuple < InitialState > > {
protected:
	std::map < std::pair < label::Label, regexp::RegExp >, std::set < label::Label > > transitions;

public:
	explicit ExtendedNFA ( label::Label initialState );
	explicit ExtendedNFA ( std::set < label::Label > states, std::set < alphabet::Symbol > inputAlphabet, label::Label initialState, std::set < label::Label > finalStates );
	explicit ExtendedNFA ( const CompactNFA < > & other );
	explicit ExtendedNFA ( const EpsilonNFA < > & other );
	explicit ExtendedNFA ( const MultiInitialStateNFA < > & other );
	explicit ExtendedNFA ( const NFA<> & other );
	explicit ExtendedNFA ( const DFA<> & other );

	virtual AutomatonBase * clone ( ) const;

	virtual AutomatonBase * plunder ( ) &&;

	const label::Label & getInitialState ( ) const {
		return accessElement < InitialState > ( ).get ( );
	}

	bool setInitialState ( label::Label state ) {
		return accessElement < InitialState > ( ).set ( std::move ( state ) );
	}

	const std::set < label::Label > & getStates ( ) const {
		return accessComponent < States > ( ).get ( );
	}

	bool addState ( label::Label state ) {
		return accessComponent < States > ( ).add ( std::move ( state ) );
	}

	void setStates ( std::set < label::Label > states ) {
		accessComponent < States > ( ).set ( std::move ( states ) );
	}

	void removeState ( const label::Label & state ) {
		accessComponent < States > ( ).remove ( state );
	}

	const std::set < label::Label > & getFinalStates ( ) const {
		return accessComponent < FinalStates > ( ).get ( );
	}

	bool addFinalState ( label::Label state ) {
		return accessComponent < FinalStates > ( ).add ( std::move ( state ) );
	}

	void setFinalStates ( std::set < label::Label > states ) {
		accessComponent < FinalStates > ( ).set ( std::move ( states ) );
	}

	void removeFinalState ( const label::Label & state ) {
		accessComponent < FinalStates > ( ).remove ( state );
	}

	const std::set < alphabet::Symbol > & getInputAlphabet ( ) const {
		return accessComponent < InputAlphabet > ( ).get ( );
	}

	bool addInputSymbol ( alphabet::Symbol symbol ) {
		return accessComponent < InputAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addInputSymbols ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < InputAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setInputAlphabet ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < InputAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removeInputSymbol ( const alphabet::Symbol & symbol ) {
		accessComponent < InputAlphabet > ( ).remove ( symbol );
	}

	/**
	 * Adds transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addTransition ( label::Label current, regexp::RegExp input, label::Label next );

	/**
	 * Removes transition from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeTransition ( const label::Label & current, const regexp::RegExp & input, const label::Label & next );

	/**
	 * @return automaton transitions
	 */
	const std::map < std::pair < label::Label, regexp::RegExp >, std::set < label::Label > > & getTransitions ( ) const;

	/**
	 * @return automaton transitions from state
	 */
	std::map < std::pair < label::Label, regexp::RegExp >, std::set < label::Label > > getTransitionsFromState ( const label::Label & from ) const;

	/**
	 * @return automaton transitions to state
	 */
	std::map < std::pair < label::Label, regexp::RegExp >, std::set < label::Label > > getTransitionsToState ( const label::Label & from ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const ExtendedNFA & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "ExtendedNFA";

		return xmlTagName;
	}

	static ExtendedNFA parse ( std::deque < sax::Token >::iterator & input );
	static void parseTransition ( std::deque < sax::Token >::iterator & input, ExtendedNFA & automaton );

	void compose ( std::deque < sax::Token > & out ) const;
	void composeTransitions ( std::deque < sax::Token > & out ) const;
};

} /* namespace automaton */

namespace std {

template < >
class ComponentConstraint< automaton::ExtendedNFA, alphabet::Symbol, automaton::InputAlphabet > {
public:
	static bool used ( const automaton::ExtendedNFA & automaton, const alphabet::Symbol & symbol ) {
		for ( const std::pair < const std::pair < label::Label, regexp::RegExp >, std::set < label::Label > > & transition : automaton.getTransitions ( ) )
			if ( transition.first.second.getAlphabet ( ).count ( symbol ) == 1 )
				return true;

		return false;
	}

	static bool available ( const automaton::ExtendedNFA &, const alphabet::Symbol & ) {
		return true;
	}

	static void valid ( const automaton::ExtendedNFA &, const alphabet::Symbol & ) {
	}
};

template < >
class ComponentConstraint< automaton::ExtendedNFA, label::Label, automaton::States > {
public:
	static bool used ( const automaton::ExtendedNFA & automaton, const label::Label & state ) {
		if ( automaton.getInitialState ( ) == state )
			return true;

		if ( automaton.getFinalStates ( ).count ( state ) )
			return true;

		for ( const std::pair < const std::pair < label::Label, regexp::RegExp >, std::set < label::Label > > & transition : automaton.getTransitions ( ) )
			if ( ( transition.first.first == state ) || ( transition.second.find ( state ) != transition.second.end ( ) ) )
				return true;

		return false;
	}

	static bool available ( const automaton::ExtendedNFA &, const label::Label & ) {
		return true;
	}

	static void valid ( const automaton::ExtendedNFA &, const label::Label & ) {
	}
};

template < >
class ComponentConstraint< automaton::ExtendedNFA, label::Label, automaton::FinalStates > {
public:
	static bool used ( const automaton::ExtendedNFA &, const label::Label & ) {
	return false;
}

	static bool available ( const automaton::ExtendedNFA & automaton, const label::Label & state ) {
		return automaton.accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::ExtendedNFA &, const label::Label & ) {
	}
};

template < >
class ElementConstraint< automaton::ExtendedNFA, label::Label, automaton::InitialState > {
public:
	static bool available ( const automaton::ExtendedNFA & automaton, const label::Label & state ) {
		return automaton.accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::ExtendedNFA &, const label::Label & ) {
	}
};

} /* namespace std */

#endif /* EXTENDED_NFA_H_ */
