#ifndef FORMAL_RTE_H_
#define FORMAL_RTE_H_

#include "../RTEBase.h"
#include <exception/CommonException.h>
#include "FormalRTEElement.h"
#include <core/components.hpp>
#include <set>
#include <string>

namespace rte {

class FormalRTEElement;
class GeneralAlphabet;
class ConstantAlphabet;

/**
 * Represents regular tree expression parsed from the XML. Regular tree expression is stored
 * as a tree of RTEElement elements.
 */
class FormalRTE : public RTEBase, public std::Components < FormalRTE, std::ranked_symbol < >, std::tuple < GeneralAlphabet, ConstantAlphabet >, std::tuple < > > {
protected:
	std::smart_ptr < FormalRTEElement > rte;

public:
	explicit FormalRTE ( );

	explicit FormalRTE ( std::set < std::ranked_symbol < > > alphabetF, std::set < std::ranked_symbol < > > alphabetK, const FormalRTEElement & rte );
	explicit FormalRTE ( std::set < std::ranked_symbol < > > alphabetF, std::set < std::ranked_symbol < > > alphabetK, FormalRTEElement && rte );
	explicit FormalRTE ( std::pair < std::set < std::ranked_symbol < > >, std::set < std::ranked_symbol < > > > alphabets, const FormalRTEElement & rte );
	explicit FormalRTE ( std::pair < std::set < std::ranked_symbol < > >, std::set < std::ranked_symbol < > > > alphabets, FormalRTEElement && rte );
	explicit FormalRTE ( const FormalRTEElement & rte );
	explicit FormalRTE ( FormalRTEElement && rte );

	/**
	 * @copydoc FormalRTEElement::clone() const
	 */
	virtual RTEBase * clone ( ) const;

	/**
	 * @copydoc FormalRTEElement::plunder() const
	 */
	virtual RTEBase * plunder ( ) &&;

	const std::set < std::ranked_symbol < > > & getAlphabet ( ) const {
		return accessComponent < GeneralAlphabet > ( ).get ( );
	}

	void addAlphabetSymbols ( std::ranked_symbol < > symbol ) {
		accessComponent < GeneralAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addAlphabetSymbols ( std::set < std::ranked_symbol < > > symbols ) {
		accessComponent < GeneralAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setAlphabetSymbols ( std::set < std::ranked_symbol < > > symbols ) {
		accessComponent < GeneralAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removeAlphabetSymbol ( const std::ranked_symbol < > & symbol ) {
		accessComponent < GeneralAlphabet > ( ).remove ( symbol );
	}

	const std::set < std::ranked_symbol < > > & getSubstitutionAlphabet ( ) const {
		return accessComponent < ConstantAlphabet > ( ).get ( );
	}

	void addConstantSymbol ( std::ranked_symbol < > symbol ) {
		accessComponent < ConstantAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addConstantSymbols ( std::set < std::ranked_symbol < > > symbols ) {
		accessComponent < ConstantAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setConstantSymbols ( std::set < std::ranked_symbol < > > symbols ) {
		accessComponent < ConstantAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removeConstantSymbol ( const std::ranked_symbol < > & symbol ) {
		accessComponent < ConstantAlphabet > ( ).remove ( symbol );
	}

	/**
	 * @return Root node of the regular expression tree
	 */
	const FormalRTEElement & getRTE ( ) const;

	/**
	 * @return Root node of the regular expression tree
	 */
	FormalRTEElement & getRTE ( );

	/**
	 * Sets the root node of the regular expression tree
	 * @param regExp root node to set
	 */
	void setRTE ( const FormalRTEElement & regExp );

	/**
	 * Sets the root node of the regular expression tree
	 * @param regExp root node to set
	 */
	void setRTE ( FormalRTEElement && regExp );

	/**
	 * Prints XML representation of the RTE to the output stream.
	 * @param out output stream to which print the RTE
	 * @param regexp RTE to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const FormalRTE & other ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName ( ) {
		static std::string xmlTagName = "FormalRTE";

		return xmlTagName;
	}

	static FormalRTE parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;
};

} /* namespace rte */

namespace std {

template < >
class ComponentConstraint< rte::FormalRTE, std::ranked_symbol < >, rte::GeneralAlphabet > {
public:
	static bool used ( const rte::FormalRTE & rte, const std::ranked_symbol < > & symbol ) {
		return rte.getRTE ( ).testSymbol ( symbol );
	}

	static bool available ( const rte::FormalRTE &, const std::ranked_symbol < > & ) {
		return true;
	}

	static void valid ( const rte::FormalRTE & rte, const std::ranked_symbol < > & symbol ) {
		if ( rte.accessComponent < rte::ConstantAlphabet > ( ).get ( ).count ( symbol ) )
			throw ::exception::CommonException ( "Symbol " + std::to_string ( symbol ) + "cannot be in general alphabet since it is already in constant alphabet" );
	}
};

template < >
class ComponentConstraint< rte::FormalRTE, std::ranked_symbol < >, rte::ConstantAlphabet > {
public:
	static bool used ( const rte::FormalRTE & rte, const std::ranked_symbol < > & symbol ) {
		return rte.getRTE ( ).testSymbol ( symbol );
	}

	static bool available ( const rte::FormalRTE &, const std::ranked_symbol < > & ) {
		return true;
	}

	static void valid ( const rte::FormalRTE & rte, const std::ranked_symbol < > & symbol ) {
		if ( rte.accessComponent < rte::GeneralAlphabet > ( ).get ( ).count ( symbol ) )
			throw ::exception::CommonException ( "Symbol " + std::to_string ( symbol ) + "cannot be in constant alphabet since it is already in general alphabet" );
	}
};

} /* namespace std */

#endif /* FORMAL_RTE_H_ */
