/*
 * RegExpFromStringParser.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "RegExpFromStringParser.h"
#include "exception/CommonException.h"
#include "regexp/RegExpClasses.h"

#include "regexp/unbounded/UnboundedRegExpElements.h"
#include "regexp/formal/FormalRegExpElements.h"
#include <regexp/RegExp.h>

#include "../StringApi.hpp"

namespace regexp {

RegExp RegExpFromStringParser::parseRegExp(std::istream& input) {
	return parseRegExp(input, std::set<FEATURES>({FEATURES::FORMAL, FEATURES::UNBOUNDED}));
}

RegExp RegExpFromStringParser::parseRegExp(std::istream& input, const std::set<FEATURES>& features) {
	std::rvalue_ref < UnboundedRegExpElement > element = alternation(input);
	UnboundedRegExp regexp(std::move(element));

	if(features.count(FEATURES::UNBOUNDED)) return RegExp{regexp};

	if(features.count(FEATURES::FORMAL)) return RegExp{FormalRegExp{regexp}};

	throw exception::CommonException("Invalid input");
}

std::rvalue_ref < UnboundedRegExpElement > RegExpFromStringParser::alternation(std::istream& input) {
	return alternationCont(input, concatenation(input));
}

std::rvalue_ref < UnboundedRegExpElement > RegExpFromStringParser::alternationCont(std::istream& input, std::rvalue_ref < UnboundedRegExpElement > left) {
	RegExpFromStringLexer::Token token = RegExpFromStringLexer::next(input);
	if(token.type == RegExpFromStringLexer::TokenType::PLUS) {
		UnboundedRegExpAlternation res;
		res.appendElement(std::move(left));
		res.appendElement(concatenation(input));

		return alternationContCont(input, std::move ( res ) );
	} else {
		RegExpFromStringLexer::putback(input, token);
		return left;
	}
}

std::rvalue_ref < UnboundedRegExpElement > RegExpFromStringParser::alternationContCont(std::istream& input, UnboundedRegExpAlternation res) {
	RegExpFromStringLexer::Token token = RegExpFromStringLexer::next(input);
	if(token.type == RegExpFromStringLexer::TokenType::PLUS) {
		res.appendElement(concatenation(input));

		return alternationContCont(input, std::move ( res ) );
	} else {
		RegExpFromStringLexer::putback(input, token);
		return std::rvalue_ref < UnboundedRegExpElement > ( std::move ( res ).plunder() );
	}
}

std::rvalue_ref < UnboundedRegExpElement > RegExpFromStringParser::concatenation(std::istream& input) {
	return concatenationCont(input, factor(input));
}

std::rvalue_ref < UnboundedRegExpElement > RegExpFromStringParser::concatenationCont(std::istream& input, std::rvalue_ref < UnboundedRegExpElement > left) {
	RegExpFromStringLexer::Token token = RegExpFromStringLexer::next(input);
	if(token.type == RegExpFromStringLexer::TokenType::ERROR || token.type == RegExpFromStringLexer::TokenType::LPAR || token.type == RegExpFromStringLexer::TokenType::EPS || token.type == RegExpFromStringLexer::TokenType::EMPTY) {
		RegExpFromStringLexer::putback(input, token);
		UnboundedRegExpConcatenation res;
		res.appendElement(std::move(left));
		res.appendElement(factor(input));

		return concatenationContCont(input, std::move ( res ) );
	} else {
		RegExpFromStringLexer::putback(input, token);
		return left;
	}
}

std::rvalue_ref < UnboundedRegExpElement > RegExpFromStringParser::concatenationContCont(std::istream& input, UnboundedRegExpConcatenation res) {
	RegExpFromStringLexer::Token token = RegExpFromStringLexer::next(input);
	if(token.type == RegExpFromStringLexer::TokenType::ERROR || token.type == RegExpFromStringLexer::TokenType::LPAR || token.type == RegExpFromStringLexer::TokenType::EPS || token.type == RegExpFromStringLexer::TokenType::EMPTY) {
		RegExpFromStringLexer::putback(input, token);
		res.appendElement(factor(input));

		return concatenationContCont(input, std::move ( res ) );
	} else {
		RegExpFromStringLexer::putback(input, token);
		return std::rvalue_ref < UnboundedRegExpElement > ( std::move ( res ).plunder() );
	}
}

std::rvalue_ref < UnboundedRegExpElement > RegExpFromStringParser::factor(std::istream& input) {
	RegExpFromStringLexer::Token token = RegExpFromStringLexer::next(input);
	if(token.type == RegExpFromStringLexer::TokenType::LPAR) {
		std::rvalue_ref < UnboundedRegExpElement > base = alternation(input);
		token = RegExpFromStringLexer::next(input);
		if(token.type != RegExpFromStringLexer::TokenType::RPAR) throw exception::CommonException("Expected RPAR");
		return star(input, std::move ( base ) );
	} else if(token.type == RegExpFromStringLexer::TokenType::EPS) {
		return star(input, std::rvalue_ref < UnboundedRegExpElement > ( new UnboundedRegExpEpsilon ( ) ) );
	} else if(token.type == RegExpFromStringLexer::TokenType::EMPTY) {
		return star(input, std::rvalue_ref < UnboundedRegExpElement > ( new UnboundedRegExpEmpty ( ) ) );
	} else if(token.type == RegExpFromStringLexer::TokenType::ERROR) {
		UnboundedRegExpSymbol res(alib::stringApi<alphabet::Symbol>::parse(input));
		return star(input, std::move ( res ) );
	} else {
		throw exception::CommonException("Unrecognised token at factor rule");
	}
}

std::rvalue_ref < UnboundedRegExpElement > RegExpFromStringParser::star(std::istream& input, UnboundedRegExpElement && elem) {
	RegExpFromStringLexer::Token token = RegExpFromStringLexer::next(input);
	if(token.type == RegExpFromStringLexer::TokenType::STAR) {
		UnboundedRegExpIteration iter ( std::move ( elem ) );
		return star(input, std::move ( iter ) );
	} else {
		RegExpFromStringLexer::putback(input, token);
		return std::rvalue_ref < UnboundedRegExpElement > ( std::move ( elem ).plunder() );
	}
}

}
